(*
Author:  Christina Kohl <christina.kohl@uibk.ac.at> (2023)
License: LGPL (see file COPYING.LESSER)
*)

text\<open>Check functions for confluence and commutation via development closed critical pairs.
Analogous to Parallel_Closed_Impl.thy\<close>

theory Development_Closed_Impl
imports
  Development_Closed
  Critical_Pair_Closure_Impl
  Commutation
begin

fun is_mstep_join where
  "is_mstep_join R S (Some n) s t =
    (\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (reachable_terms S t n)))"
| "is_mstep_join R S None s t =
    (\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (mstep_rewrite S t)))"


lemma is_mstep_join[dest]: assumes "is_mstep_join R S hint s t" 
  and varcond: "list_all (\<lambda>(l,r). is_Fun l \<and> vars_term r \<subseteq> vars_term l) S" 
  shows "\<exists>v. is_mstep R s v \<and> (t, v) \<in> (rstep (set S))\<^sup>*" 
proof (cases hint)
  case (Some n)
  let ?find = "\<lambda> v. is_mstep R s v" 
  from Some assms have "\<not> Option.is_none (List.find ?find (reachable_terms S t n))"  
    by auto
  then obtain v where "List.find ?find (reachable_terms S t n) = Some v"
    by force
  then have "is_mstep R s v" and "v \<in> set (reachable_terms S t n)"
    unfolding find_Some_iff by auto
  then show ?thesis
    using reachable_terms by blast
next
  case None
  let ?find = "\<lambda> v. is_mstep R s v"  
  from None assms have "\<not> Option.is_none (List.find ?find (mstep_rewrite S t))"
    by auto
  then obtain v where  "List.find ?find (mstep_rewrite S t) = Some v"
    by force
  then have "is_mstep R s v" and "v \<in> set (mstep_rewrite S t)"
    unfolding find_Some_iff by auto
  then show ?thesis using mstep_rsteps_subset mstep_rewrite_mstep[OF varcond]
    by blast  
qed

declare is_mstep_join.simps[simp del]

definition check_development_closed_comm where
  "check_development_closed_comm R S n = do {
     check_wf_trs R;
     check_wf_trs S;
     check_left_linear_trs R;
     check_left_linear_trs S;
     check_allm (\<lambda> (b, s, t). do {
       check (is_mstep_join S R n s t)
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' R<- . ->S '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not almost development closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl S R);
     check_allm (\<lambda> (b, s, t). do {
       check (b \<or> is_mstep R s t)
         (showsl_lit (STR ''the inner critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' S<- . ->R '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not closed with multi-step over R'') \<circ> showsl n)
     }) (critical_pairs_impl R S)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the almost development closed check for the following TRSs could not be proven\<newline>R: '') \<circ> 
        showsl_trs R o showsl_lit (STR ''\<newline>\<newline>S: '') o showsl_trs S)"

lemma check_development_closed_comm:
  assumes "isOK(check_development_closed_comm R S n)"
  shows "sig_commute F (set R) (set S)"
proof (intro commute_imp_sig_commute mstep_closed_imp_commute)
  note assms = assms[unfolded check_development_closed_comm_def, simplified, unfolded wf_trs_def']
  from assms show "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
    by auto 
  from assms show "\<And>l r. (l, r) \<in> set S \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
    by auto 
  from assms show "left_linear_trs (set R)" "left_linear_trs (set S)" by auto
  from assms show "(b, p, q) \<in> critical_pairs (set S) (set R) \<Longrightarrow> \<exists>v. (p, v) \<in> mstep (set S) \<and> (q, v) \<in> (rstep (set R))\<^sup>*" for b p q
    using is_mstep_join by (metis (mono_tags, lifting) Ball_set_list_all case_prodD is_mstep) 
  from assms show "\<And>p q. (False, p, q) \<in> critical_pairs (set R) (set S) \<Longrightarrow> (p,q) \<in> mstep (set R)"
    using is_mstep by (smt (verit) Ball_set_list_all case_prodD) 
qed

end
