(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009, 2010)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Examples
imports "../TermFun" Loga
begin

section \<open>Example - An Odd Even/Odd Function\<close>

text \<open>The first argument is a flag where @{term "True"} indicates that we
want to check for even numbers and @{term "False"} that we want to check for
odd numbers.\<close>
function evenodd :: "bool \<Rightarrow> nat \<Rightarrow> bool"
where
  "evenodd True 0 = False"
| "evenodd False x = (\<not> evenodd True x)"
| "evenodd True (Suc x) = evenodd False x"
  by (pat_completeness) auto

text \<open>As in @{theory Loga}, we first generate a TRS representation of the
involved functions. For logical negation @{const "Not"}, we use two equations
from the standard library.\<close>
define_trs not_True_eq_False not_False_eq_True (* order is important, why? *)
define_trs (no_sim) evenodd.psimps

text \<open>The resulting TRS has the name @{term "evenodd.trs"} and its definition
is provided via the lemma @{thm[source] evenodd.trs_def}.\<close>
term evenodd.trs
thm evenodd.trs_def

text \<open>The termination proof involves the same steps as for @{theory Loga}.\<close>
termination evenodd
  txt \<open>First we reduce termination of the function to termination of the TRS.\<close>
  apply (trs_reduce evenodd)
  txt \<open>Then we obtain a representation of the TRS as finite list of rules.\<close>
  apply expand_trs_problem
  txt \<open>We export the resulting list of rules into an XML file.\<close>
  export_termination_problem "evenodd.xml"
  txt \<open>A corresponding XML certificate, produced by some external termination tool,
  is reconstructed\<close>
  apply (ceta_import_proof "proofs/evenodd.proof.xml")
  txt \<open>and checked.\<close>
  apply eval
done


section \<open>Example - Normalization of Conditionals (cf. Paulson 1986)\<close>

datatype exp = AT nat | IF exp exp exp

datatype_compat exp

text \<open>Since a user-defined datatype is involved, we have to register it in order
to obtain an embedding function @{const "embed"} encoding values of the datatype by
first-order terms.\<close>
setup \<open>Term_Fun.register_datatypes [@{type_name exp}]\<close>
thm embed_exp.simps

function norm where
  "norm (AT a) = AT a"
| "norm (IF (AT a) y z) = IF (AT a) (norm y) (norm z)"
| "norm (IF (IF u v w) y z) = norm (IF u (IF v y z) (IF w y z))"
  by pat_completeness auto

text \<open>A TRS representation of @{const norm} is established using the
@{text define_trs} command.\<close>
define_trs (no_sim) norm.psimps

termination norm 
  apply (trs_reduce norm)
  apply expand_trs_problem
  export_termination_problem "norm.xml"
  apply (ceta_import_proof "proofs/norm.proof.xml")
  apply eval
done


section \<open>Example - Preprocessing of Formulas (Chaieb)\<close>

datatype num = C int | Bound nat | CN nat int num | Neg num | Add num num| Sub num num 
  | Mul int num | Floor num| CF int num num

datatype_compat num

text \<open>Obtain a term representation of the type @{typ "num"}\<close>
setup \<open>Term_Fun.register_datatypes [@{type_name num}]\<close>

text \<open>since int is not embedded, we will see dummy terms in embed_num.simps\<close>
thm embed_num.simps

datatype fm  = 
  T| F| Lt num| Le num| Gt num| Ge num| Eq num| NEq num| Dvd int num| NDvd int num|
  NOT fm| And fm fm|  Or fm fm| Imp fm fm| Iff fm fm| E fm| A fm

datatype_compat fm

text \<open>as well as the type @{typ "fm"}.\<close>
setup \<open>Term_Fun.register_datatypes [@{type_name fm}]\<close>
thm embed_fm.simps

function (sequential) prep :: "fm \<Rightarrow> fm" where
  "prep (E T) = T"
| "prep (E F) = F"
| "prep (E (Or p q)) = Or (prep (E p)) (prep (E q))"
| "prep (E (Imp p q)) = Or (prep (E (NOT p))) (prep (E q))"
| "prep (E (Iff p q)) = Or (prep (E (And p q))) (prep (E (And (NOT p) (NOT q))))" 
| "prep (E (NOT (And p q))) = Or (prep (E (NOT p))) (prep (E(NOT q)))"
| "prep (E (NOT (Imp p q))) = prep (E (And p (NOT q)))"
| "prep (E (NOT (Iff p q))) = Or (prep (E (And p (NOT q)))) (prep (E(And (NOT p) q)))"
| "prep (E p) = E (prep p)"
| "prep (A (And p q)) = And (prep (A p)) (prep (A q))"
| "prep (A p) = prep (NOT (E (NOT p)))"
| "prep (NOT (NOT p)) = prep p"
| "prep (NOT (And p q)) = Or (prep (NOT p)) (prep (NOT q))"
| "prep (NOT (A p)) = prep (E (NOT p))"
| "prep (NOT (Or p q)) = And (prep (NOT p)) (prep (NOT q))"
| "prep (NOT (Imp p q)) = And (prep p) (prep (NOT q))"
| "prep (NOT (Iff p q)) = Or (prep (And p (NOT q))) (prep (And (NOT p) q))"
| "prep (NOT p) = NOT (prep p)"
| "prep (Or p q) = Or (prep p) (prep q)"
| "prep (And p q) = And (prep p) (prep q)"
| "prep (Imp p q) = prep (Or (NOT p) q)"
| "prep (Iff p q) = Or (prep (And p q)) (prep (And (NOT p) (NOT q)))"
| "prep p = p"
  by pat_completeness simp_all

text \<open>The remaining steps are ``canonical.''\<close>
define_trs (no_sim) prep.psimps

termination prep
  apply (trs_reduce prep)
  apply expand_trs_problem
  export_termination_problem "prep.xml"
  apply (ceta_import_proof "proofs/prep.proof.xml")
  apply eval
done


section \<open>Example - Merge Sort\<close>

fun merge :: "nat list \<Rightarrow> nat list \<Rightarrow> nat list" where
  "merge (x#xs) (y#ys) = (case x \<le> y of
    True => x # merge xs (y#ys)
  |  False => y # merge (x#xs) ys)"
| "merge xs [] = xs"
| "merge [] ys = ys"
(*  by pat_completeness auto *)

function (sequential) msort :: "nat list \<Rightarrow> nat list" where
  "msort [] = []"
| "msort [x] = [x]"
| "msort xs = merge (msort (take (half (length xs)) xs))
	            (msort (drop (half (length xs)) xs))"
  by pat_completeness auto

(*
define_trs (no_sim) merge.psimps

text {* For @{const "merge"}, we are able to prove termination using our approach. *}
termination merge
  apply (trs_reduce merge)
  apply expand_trs_problem
  export_termination_problem "merge.xml"
  apply (ceta_import_proof "proofs/merge.proof.xml")
  apply eval
done
*)

define_trs merge.simps

define_trs (no_sim) msort.psimps
thm msort.trs_def

termination msort
  apply (trs_reduce msort)
  apply expand_trs_problem
  export_termination_problem "msort.xml"
  apply (ceta_import_proof "proofs/msort.proof.xml")
  apply eval
done

end
